package cn.iocoder.yudao.module.system.service.robot;

import cn.hutool.core.util.ObjectUtil;
import cn.iocoder.yudao.framework.common.util.date.DateUtils;
import cn.iocoder.yudao.module.system.controller.admin.robot.vo.RobotTaskDetailAddVo;
import cn.iocoder.yudao.module.system.controller.admin.robot.vo.RobotTaskPageReqVO;
import cn.iocoder.yudao.module.system.controller.admin.robot.vo.RobotTaskSaveReqVO;
import cn.iocoder.yudao.module.system.dal.dataobject.houselocation.WareHouseLocationDO;
import cn.iocoder.yudao.module.system.dal.dataobject.robot.RobotTaskDO;
import cn.iocoder.yudao.module.system.dal.dataobject.robot.RobotTaskDetailDO;
import cn.iocoder.yudao.module.system.dal.mysql.houselocation.WareHouseLocationMapper;
import cn.iocoder.yudao.module.system.dal.mysql.robot.RobotTaskDetailMapper;
import cn.iocoder.yudao.module.system.dal.mysql.robot.RobotTaskMapper;
import cn.iocoder.yudao.module.system.enums.redis.RobotCacheTypeEnum;
import cn.iocoder.yudao.module.system.enums.robot.*;
import cn.iocoder.yudao.module.system.util.redis.RedisUtil;
import cn.iocoder.yudao.module.system.util.redis.RedissonUtils;
import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.redisson.api.RLock;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import javax.annotation.Resource;

import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.common.util.object.BeanUtils;


import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception0;
import static cn.iocoder.yudao.module.system.enums.ErrorCodeConstants.*;

/**
 * 机器人任务主表 Service 实现类
 *
 * @author 陈宾顺
 */
@Slf4j
@Service
@Validated
public class RobotTaskServiceImpl implements RobotTaskService {

    private final String RESULT = "SUCCESS";

    @Resource
    private RobotTaskMapper taskMapper;

    @Resource
    private WareHouseLocationMapper locationMapper;

    @Resource
    private RedisUtil redisUtil;

    @Resource
    private RedissonUtils redissonUtils;

    @Resource
    private RobotTaskDetailMapper taskDetailMapper;

    @Resource
    private RobotTaskDetailService robotTaskDetailService;

    @Value("${zn.task-no:ZN}")
    private String taskNo;
    @Value("${zn.do_cycle:true}")
    private Boolean doCycle;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Long createTask(RobotTaskSaveReqVO createReqVO) throws InterruptedException {
        //校验
        doTaskCheck(createReqVO);

        //设置任务号
        if (ObjectUtil.isEmpty(createReqVO.getTaskNo())) {
            String incrementByKey = redisUtil.getIncrementByKey(RobotCacheTypeEnum.TASK_NO.getKey());
            createReqVO.setTaskNo(taskNo+ DateUtils.getYearMonthDay()+incrementByKey);
        }
        //获取库位id
        RLock lock = redissonUtils.getLock(RobotCacheTypeEnum.ROBOT_TASK_ADD_LOCK.getKey());

        String addResult = "";
        if (lock.tryLock(30l,TimeUnit.MINUTES)){
            try {
                addResult = addTask(createReqVO);
            } catch (Exception e) {
                log.error("下发任务出现异常 :{}",e);
                addResult = e.getMessage();
            } finally {
                lock.unlock();
            }
        }else {
            throw exception(REDISSON_NOT_OBTAIN_LOCK);
        }

        if (!RESULT.equals(addResult)) {
            throw exception0(TASK_CHECK_EXCEPTION.getCode(), addResult);
        }

        return 1l;
    }

    /**
     * 校验
     * @param createReqVO
     */
    private void doTaskCheck(RobotTaskSaveReqVO createReqVO) {
        //校验明细
        if (ObjectUtil.isEmpty(createReqVO.getTaskDetailList())) {
            throw exception(TASK_DATA_NOT_FULL);
        }
        //校验循环任务的配置未开启
        if (DoCycleEnum.YES.getType().equals(createReqVO.getDoCycle()) && !doCycle) {
            throw exception(TASK_CYCLE_NOT_OPEN);
        }
        //搬空任务只能选择线库或者区域
        if (MoveAllEnum.YES.getType().equals(createReqVO.getDoMoveAll())) {
            Set<Integer> releaseTypeList =
                    createReqVO.getTaskDetailList().stream().map(v -> v.getReleaseType())
                            .collect(Collectors.toSet());
            Set<Integer> takeTypeList =
                    createReqVO.getTaskDetailList().stream().map(v -> v.getTakeType())
                            .collect(Collectors.toSet());
            if (releaseTypeList.contains(ReleaseTakeEnum.TO_LOCATION.getType())
                    || takeTypeList.contains(ReleaseTakeEnum.TO_LOCATION.getType())) {
                throw exception(TASK_ONLY_CHOOSE_LOCATION);
            }
        }
    }

    /**
     * 添加任务
     * @param createReqVO
     */
    @Transactional(rollbackFor = Exception.class)
    public String addTask(RobotTaskSaveReqVO createReqVO) {
        RobotTaskDO task = BeanUtils.toBean(createReqVO, RobotTaskDO.class);
        taskMapper.insert(task);

        //查找库位
        if (MoveAllEnum.NO.getType().equals(createReqVO.getDoMoveAll())) {
            setSingleLocationIdNo(createReqVO.getTaskDetailList(),task.getId());
        } else {
            setAllLocationIdNo(createReqVO,task.getId());
        }

        // 插入
        createReqVO.getTaskDetailList()
                .forEach(taskDetail -> {
                    taskDetail.setRobotTaskId(task.getId());
                });
        robotTaskDetailService.insertRobotDetailList(createReqVO.getTaskDetailList());

        return RESULT;
    }

    /**
     * 设置放货和取货的id
     * @param createReqVO
     * @param taskId
     */
    @Transactional(rollbackFor = Exception.class)
    public void setAllLocationIdNo(RobotTaskSaveReqVO createReqVO, Long taskId) {
        List<RobotTaskDetailAddVo> taskDetailList = createReqVO.getTaskDetailList();

        List<RobotTaskDetailAddVo> newTaskDetailList = new ArrayList<>();
        List<Long> toLocationIds = new ArrayList<>(); //被此次任务锁定的空库位id
        List<Long> fromLocationIds = new ArrayList<>(); //被此次任务锁定的有货库位id
        for (RobotTaskDetailAddVo robotTaskVo : taskDetailList) {
            //查找有货且非锁定的库位
            WareHouseLocationDO query = null;
            if (ReleaseTakeEnum.TO_LANE.getType().equals(robotTaskVo.getTakeType())) {
                query = WareHouseLocationDO.builder()
                        .laneId(robotTaskVo.getTakeId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.YES.getType())
                        .build();
            } else if (ReleaseTakeEnum.TO_AREA.getType().equals(robotTaskVo.getTakeType())) {
                query = WareHouseLocationDO.builder()
                        .areaId(robotTaskVo.getTakeId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.YES.getType())
                        .build();
            }

            List<WareHouseLocationDO> stockList = locationMapper.selectLocations(query,toLocationIds);
            if (ObjectUtil.isEmpty(stockList)) {
                throw new RuntimeException("取货线库/区域为空或者已锁定");
            }
            //判断取货库位是否存在未完成的任务
            Set<Long> stockLocationIds = stockList.stream().map(WareHouseLocationDO::getId).collect(Collectors.toSet());
            List<RobotTaskDetailDO> doingTasks = taskDetailMapper.queryDoingTaskByIds(stockLocationIds);
            if (ObjectUtil.isNotEmpty(doingTasks)) {
                throw new RuntimeException("取货线库/区域,存在未完成的任务");
            }
            //查找空库位
            WareHouseLocationDO releaseQuery = null;
            if (ReleaseTakeEnum.TO_LANE.getType().equals(robotTaskVo.getReleaseType())) {
                releaseQuery = WareHouseLocationDO.builder()
                        .laneId(robotTaskVo.getReleaseId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.NO.getType())
                        .build();
            } else if (ReleaseTakeEnum.TO_AREA.getType().equals(robotTaskVo.getReleaseType())) {
                 releaseQuery = WareHouseLocationDO.builder()
                        .areaId(robotTaskVo.getReleaseId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.NO.getType())
                        .build();
            }

            List<WareHouseLocationDO> releaseStockList = locationMapper.selectLocations(releaseQuery,toLocationIds);
            if (ObjectUtil.isEmpty(releaseStockList) || releaseStockList.size() < stockList.size()) {
                throw new RuntimeException("放货线库/区域库位数量不足");
            }
            //赋值数据
            List<RobotTaskDetailAddVo> addTaskDetailList = getTaskDatail(stockList,releaseStockList,toLocationIds,
                                                                            robotTaskVo,taskId,fromLocationIds);
            newTaskDetailList.addAll(addTaskDetailList);

        }

        createReqVO.setTaskDetailList(newTaskDetailList);
        //跟新库位状态为锁定
        List<List<Long>> toPartition = Lists.partition(toLocationIds, 100);
        toPartition.forEach(list-> {
            locationMapper.updateLocationLockList(list,taskId);
        });

        List<List<Long>> fromPartition = Lists.partition(fromLocationIds, 100);
        fromPartition.forEach(list-> {
            locationMapper.updateLocationLockList(list,taskId);
        });

    }

    /**
     * 赋值数据
     * @param stockList
     * @param releaseStockList
     * @param toLocationIds
     * @param robotTaskVo
     * @return
     */
    private List<RobotTaskDetailAddVo> getTaskDatail(List<WareHouseLocationDO> stockList,
                                                     List<WareHouseLocationDO> releaseStockList,
                                                     List<Long> toLocationIds, RobotTaskDetailAddVo robotTaskVo,
                                                     Long taskId, List<Long> fromLocationIds) {
        List<RobotTaskDetailAddVo> addTaskDetailList = new ArrayList<>();
        for (int i = 0; i < stockList.size(); i++) {
            RobotTaskDetailAddVo robotTaskDetailAddVo = new RobotTaskDetailAddVo();
            robotTaskDetailAddVo.setTaskType(robotTaskVo.getTaskType());
            robotTaskDetailAddVo.setReleaseType(robotTaskVo.getReleaseType());
            robotTaskDetailAddVo.setTakeType(robotTaskVo.getTakeType());
            robotTaskDetailAddVo.setTakeId(robotTaskVo.getTakeId());
            robotTaskDetailAddVo.setReleaseId(robotTaskVo.getReleaseId());
            robotTaskDetailAddVo.setRobotNo(robotTaskVo.getRobotNo());
            robotTaskDetailAddVo.setFromLocationNo(stockList.get(i).getLocationNo());
            robotTaskDetailAddVo.setFromLocationId(stockList.get(i).getId());
            robotTaskDetailAddVo.setToLocationNo(releaseStockList.get(i).getLocationNo());
            robotTaskDetailAddVo.setToLocationId(releaseStockList.get(i).getId());
            robotTaskDetailAddVo.setRobotTaskId(taskId);
            robotTaskDetailAddVo.setPriority(robotTaskVo.getPriority());
            robotTaskDetailAddVo.setFromLocationStorey(stockList.get(i).getLocationStorey());
            robotTaskDetailAddVo.setToLocationStorey(releaseStockList.get(i).getLocationStorey());
            addTaskDetailList.add(robotTaskDetailAddVo);

            toLocationIds.add(releaseStockList.get(i).getId());
            fromLocationIds.add(stockList.get(i).getId());
        }
        return addTaskDetailList;
    }

    /**
     * 设置放货和取货的id
     * @param taskDetailList
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public void setSingleLocationIdNo(List<RobotTaskDetailAddVo> taskDetailList, Long taskId){
        List<Long> locationIds = new ArrayList<>();
        for (RobotTaskDetailAddVo robotTaskVo : taskDetailList) {
            //校验放货库位
            if (ReleaseTakeEnum.TO_LOCATION.getType().equals(robotTaskVo.getReleaseType())) {
                WareHouseLocationDO query = WareHouseLocationDO.builder().id(robotTaskVo.getReleaseId()).build();
                WareHouseLocationDO locationDO = locationMapper.queryAllByLimit(query);
                robotTaskVo.setToLocationNo(locationDO.getLocationNo());
                robotTaskVo.setToLocationId(robotTaskVo.getReleaseId());
                robotTaskVo.setToLocationStorey(locationDO.getLocationStorey());
            } else {
                WareHouseLocationDO wareHouseLocationDO =
                        locationMapper.selectByTypeAndId(LocationUseStatusEnum.NO.getType(), robotTaskVo.getReleaseType()
                                ,robotTaskVo.getReleaseId(),locationIds);
                if (ObjectUtil.isEmpty(wareHouseLocationDO)) {
                    throw new RuntimeException("放货库位为空或者已锁定 "+ robotTaskVo.getReleaseId());
                }
                robotTaskVo.setToLocationNo(wareHouseLocationDO.getLocationNo());
                robotTaskVo.setToLocationId(wareHouseLocationDO.getId());
                robotTaskVo.setToLocationStorey(wareHouseLocationDO.getLocationStorey());
            }
            robotTaskVo.setPriority(robotTaskVo.getPriority());
            locationIds.add(robotTaskVo.getToLocationId());
            //校验取货库位
            if (ReleaseTakeEnum.TO_LOCATION.getType().equals(robotTaskVo.getTakeType())) {
                WareHouseLocationDO query = WareHouseLocationDO.builder().id(robotTaskVo.getTakeId()).build();
                WareHouseLocationDO locationDO = locationMapper.queryAllByLimit(query);
                robotTaskVo.setFromLocationNo(locationDO.getLocationNo());
                robotTaskVo.setFromLocationId(robotTaskVo.getTakeId());
                robotTaskVo.setFromLocationStorey(locationDO.getLocationStorey());
            } else {
                WareHouseLocationDO wareHouseLocationDO =
                        locationMapper.selectByTypeAndId(LocationUseStatusEnum.YES.getType(),robotTaskVo.getTakeType(),
                                robotTaskVo.getTakeId(),locationIds);
                if (ObjectUtil.isEmpty(wareHouseLocationDO)) {
                    throw new RuntimeException("取货库位为空或者已锁定 "+ robotTaskVo.getTakeId());
                }
                robotTaskVo.setFromLocationNo(wareHouseLocationDO.getLocationNo());
                robotTaskVo.setFromLocationId(wareHouseLocationDO.getId());
                robotTaskVo.setFromLocationStorey(wareHouseLocationDO.getLocationStorey());
            }
            locationIds.add(robotTaskVo.getFromLocationId());
            //验证取货库位是否存在未完成的任务
            Set<Long> set  = new HashSet<>();
            set.add(robotTaskVo.getToLocationId());
            List<RobotTaskDetailDO> doingTasks = taskDetailMapper.queryDoingTaskByIds(set);
            if (ObjectUtil.isNotEmpty(doingTasks)) {
                throw new RuntimeException("取货库位,存在未完成的任务");
            }

            //设置为锁定
            locationMapper.updateLocationLockStatus(robotTaskVo.getToLocationId(),0,taskId);
            locationMapper.updateLocationLockStatus(robotTaskVo.getFromLocationId(),0,taskId);
        }
    }

    @Override
    public void updateTask(RobotTaskSaveReqVO updateReqVO) {
        // 校验存在
        validateTaskExists(updateReqVO.getId());
        // 更新
        RobotTaskDO updateObj = BeanUtils.toBean(updateReqVO, RobotTaskDO.class);
        taskMapper.updateById(updateObj);
    }

    @Override
    public void deleteTask(Long id) {
        // 校验存在
        validateTaskExists(id);
        // 删除
        taskMapper.deleteById(id);
    }

    private void validateTaskExists(Long id) {
        if (taskMapper.selectById(id) == null) {
            throw exception(TASK_NOT_EXISTS);
        }
    }

    @Override
    public RobotTaskDO getTask(Long id) {
        return taskMapper.selectById(id);
    }

    @Override
    public PageResult<RobotTaskDO> getTaskPage(RobotTaskPageReqVO pageReqVO) {
        return taskMapper.selectPage(pageReqVO);
    }

}